// Copyright (C) 2018-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package xbtree
// ΔBtail

import (
	"context"
	"fmt"
	"sort"
	"strings"

	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"

	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xbtree/blib"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xtail"
)

const traceΔBtail = false
const debugΔBtail = false


// ΔBtail represents tail of revisional changes to BTrees.
//
// It semantically consists of
//
//	[]δB			; rev ∈ (tail, head]
//
// where δB represents a change in BTrees space
//
//	δB:
//		.rev↑
//		{} root -> {}(key, δvalue)
//
// It covers only changes to keys from tracked subset of BTrees parts.
// In particular a key that was not explicitly requested to be tracked, even if
// it was changed in δZ, is not guaranteed to be present in δB.
//
// ΔBtail provides the following operations:
//
//   .Track(path)		- start tracking tree nodes and keys; root=path[0], keys=path[-1].(lo,hi]
//
//   .Update(δZ) -> δB				- update BTree δ tail given raw ZODB changes
//   .ForgetPast(revCut)			- forget changes past revCut
//   .SliceByRev(lo, hi) -> []δB		- XXX
//   .SliceByRootRev(root, lo, hi) -> []δT	- XXX
//
// XXX also:
//   .Get(root, key, at)			- get root[key] @at assuming root[key] ∈ tracked
//
// An example for tracked set is a set of visited BTree paths.
// There is no requirement that tracked set belongs to only one single BTree.
//
// XXX δB can convert δZ to btree changes, but only at least for δZ's objects
// that ∈ BTree subgraphs that were explicitly requested to be tracked by δB.
//
// XXX incremental; not full coverage
//
// ΔBtail is not safe for concurrent access.	XXX rework
// XXX -> multiple readers / single writer?
//
// See also zodb.ΔTail
type ΔBtail struct {
	// raw ZODB changes; Kept to rebuild .vδTbyRoot after new Track.
	// includes all changed objects, not only tracked ones.
	δZtail *zodb.ΔTail

	vδBroots  []_ΔBroots           // [] (rev↑, roots changed in this rev)
	vδTbyRoot map[zodb.Oid]*ΔTtail // {} root -> [] k/v change history; only for keys ∈ tracked subset	XXX -> byRoot?

	// set of tracked nodes as of @head state.
	// For this set all vδT are fully computed.
	// The set of nodes that were requested to be tracked, but were not yet
	// taken into account, is kept in ΔTtail.trackNew & co.
	trackSet blib.PPTreeSubSet

	// set of trees for which .trackNew is non-empty
	trackNewRoots setOid

	// handle to make connections to access database.
	// TODO allow client to optionally provide zconnOld/zconnNew on e.g. Update()
	db *zodb.DB    // to open connections to load new/old tree|buckets
}

// ΔTtail represent tail of revisional changes to one BTree.
//
// See ΔBtail documentation for details.
type ΔTtail struct {
	vδT []ΔTree // changes to tree keys; rev↑. covers keys ∈ tracked subset

	// set of nodes that were requested to be tracked in this tree, but for
	// which vδT was not yet rebuilt
	trackNew blib.PPTreeSubSet
	// XXX + trackNewKeys RangedKeySet

	// index for LastRevOf queries
	lastRevOf map[Key]zodb.Tid // {} key -> last
}

// _ΔBroots represents roots-only part of ΔB.
//
// It describes which trees were changed, but does not provide δkv details for changed trees.
type _ΔBroots struct {
	Rev    zodb.Tid
	ΔRoots setOid   // which roots changed in this revision
}

// ΔB represents a change in BTrees space.
type ΔB struct {
	Rev     zodb.Tid
	ΔByRoot map[zodb.Oid]map[Key]ΔValue // {} root -> {}(key, δvalue)	XXX -> ByRoot?
}

// ΔTree describes changes to one BTree in one revision.
// XXX -> ΔT ?
type ΔTree struct {
	Rev  zodb.Tid
	ΔKV  map[Key]ΔValue
}


// NewΔBtail creates new empty ΔBtail object.
//
// Initial tracked set is empty.
// Initial coverage is (at₀, at₀].
//
// db will be used by ΔBtail to open database connections to load data from
// ZODB when needed.   XXX or caller provides zhead/zprev explicitly?
func NewΔBtail(at0 zodb.Tid, db *zodb.DB) *ΔBtail {
	return &ΔBtail{
		δZtail:        zodb.NewΔTail(at0),
		vδBroots:      nil,
		vδTbyRoot:     map[zodb.Oid]*ΔTtail{},
		trackSet:      blib.PPTreeSubSet{},
		trackNewRoots: setOid{},
		db:            db,
	}
}

// newΔTtail creates new empty ΔTtail object.
func newΔTtail() *ΔTtail {
	return &ΔTtail{
		trackNew:  blib.PPTreeSubSet{},
		lastRevOf: make(map[Key]zodb.Tid),
	}
}

// Clone returns copy of ΔBtail.
func (orig *ΔBtail) Clone() *ΔBtail {
	klon := &ΔBtail{db: orig.db}

	// δZtail
	klon.δZtail = zodb.NewΔTail(orig.Tail())
	for _, δZ := range orig.δZtail.Data() {
		klon.δZtail.Append(δZ.Rev, δZ.Changev)
	}

	// vδBroots
	klon.vδBroots = make([]_ΔBroots, 0, len(orig.vδBroots))
	for _, origδBroots := range orig.vδBroots {
		klonδBroots := _ΔBroots{
			Rev:    origδBroots.Rev,
			ΔRoots: origδBroots.ΔRoots.Clone(),
		}
		klon.vδBroots = append(klon.vδBroots, klonδBroots)
	}

	// vδTbyRoot
	klon.vδTbyRoot = make(map[zodb.Oid]*ΔTtail, len(orig.vδTbyRoot))
	for root, origΔTtail := range orig.vδTbyRoot {
		klon.vδTbyRoot[root] = origΔTtail.Clone()
	}

	// trackSet, trackNewRoots
	klon.trackSet = orig.trackSet.Clone()
	klon.trackNewRoots = orig.trackNewRoots.Clone()

	return klon
}

// Clone returns copy of ΔTtail.
func (orig *ΔTtail) Clone() *ΔTtail {
	klon := &ΔTtail{}
	klon.vδT = vδTClone(orig.vδT)
	klon.trackNew = orig.trackNew.Clone()
	klon.lastRevOf = make(map[Key]zodb.Tid, len(orig.lastRevOf))
	for k, rev := range orig.lastRevOf {
		klon.lastRevOf[k] = rev
	}
	return klon
}

// vδTClone returns deep copy of []ΔTree.
func vδTClone(orig []ΔTree) []ΔTree {
	klon := make([]ΔTree, 0, len(orig))
	for _, origδT := range orig {
		klonδT := ΔTree{
			Rev: origδT.Rev,
			ΔKV: make(map[Key]ΔValue, len(origδT.ΔKV)),
		}
		for k, δv := range origδT.ΔKV {
			klonδT.ΔKV[k] = δv
		}
		klon = append(klon, klonδT)
	}
	return klon
}

// (tail, head] coverage
func (δBtail *ΔBtail) Head() zodb.Tid { return δBtail.δZtail.Head() }
func (δBtail *ΔBtail) Tail() zodb.Tid { return δBtail.δZtail.Tail() }

// Track adds tree path to tracked set.
//
// path[0] signifies tree root.
// All path elements must be Tree except last one which must be Bucket.
//
// Besides key (which might point to value or hole), δBtail will also track all
// keys residing in tracked leaf nodes. In particular after request for KeyMax or
// KeyMin to be tracked, δBtail will keep on tracking changes to maximum or
// minimum keys correspondingly.
//
// XXX objects in path must be with .PJar().At() == .head
// XXX path -> []oid ?
//
// XXX catch cycles on add?
// XXX no need to pass in key?  (-> all keys, covered by leaf keyrange, will be added to tracking set of keys)
// XXX no need to return error?
func (δBtail *ΔBtail) Track(key Key, nodePath []Node) error { // XXX Tree|Bucket; path[0] = root

	path := nodePathToPath(nodePath)

	pathv := []string{}
	for _, node := range nodePath { pathv = append(pathv, vnode(node)) }
	tracefΔBtail("\nTrack [%v] %s\n", key, strings.Join(pathv, " -> "))
	tracefΔBtail("trackSet: %s\n", δBtail.trackSet) // XXX locking

	return δBtail.track(key, path)
}

func (δBtail *ΔBtail) track(key Key, path []zodb.Oid) error {
	// XXX locking

	// first normalize path: remove embedded bucket and check if it was an
	// empty artificial tree. We need to do the normalization because we
	// later check whether leaf path[-1] ∈ trackSet and without
	// normalization path[-1] can be InvalidOid.
	path = blib.NormPath(path)
	if len(path) == 0 {
		return nil // empty tree
	}

	root := path[0]

	// nothing to do if key is already tracked
	leaf := path[len(path)-1]
	if δBtail.trackSet.Has(leaf) {
		tracefΔBtail("->T: nop\n")
		path_ := δBtail.trackSet.Path(leaf)
		if !pathEqual(path, path_) {
			panicf("BUG: key %s is already tracked via path=%v\ntrack requests path=%v",  kstr(key), path_, path)
		}
		return nil
	}

	// queue path into trackNew
	δTtail, ok := δBtail.vδTbyRoot[root]
	if !ok {
		δTtail = newΔTtail()
		δBtail.vδTbyRoot[root] = δTtail
	}

	δBtail.trackNewRoots.Add(root)
	δTtail.trackNew.AddPath(path)
	tracefΔBtail("->T: [%s].trackNew -> %s\n", root, δTtail.trackNew)
	return nil
}

// rebuildAll rebuilds ΔBtail taking all trackNew requests into account.
func (δBtail *ΔBtail) rebuildAll() (err error) {
	defer xerr.Context(&err, "ΔBtail rebuildAll")
	// XXX locking

	tracefΔBtail("\nRebuildAll @%s..@%s    trackNewRoots: %s\n", δBtail.Tail(), δBtail.Head(), δBtail.trackNewRoots)

	for root := range δBtail.trackNewRoots {
		delete(δBtail.trackNewRoots, root)
		δBtail.rebuild1(root)
	}

	return nil
}

// rebuild1IfNeeded rebuilds ΔBtail for single root if that root needs rebuilding.
func (δBtail *ΔBtail) rebuild1IfNeeded(root zodb.Oid) error {
	// XXX locking
	_, ok := δBtail.trackNewRoots[root]
	if !ok {
		return nil
	}

	delete(δBtail.trackNewRoots, root)
	return δBtail.rebuild1(root)
}

// rebuild1 rebuilds ΔBtail for single root.
func (δBtail *ΔBtail) rebuild1(root zodb.Oid) error {
	// XXX locking
	δTtail := δBtail.vδTbyRoot[root] // must be there
	δtrackSet, δrevSet, err := δTtail.rebuild(root, δBtail.δZtail, δBtail.db)
	if err != nil {
		return err
	}
	δBtail.trackSet.UnionInplace(δtrackSet)
	δBtail.vδBroots_Update(root, δrevSet)
	return nil
}


// rebuild rebuilds ΔTtail taking trackNew requests into account.
//
// It returns:
//
// - set of nodes that must be added to ΔBtail.trackSet to account for
//   keys that becomes tracked. Note: this set is potentially wider compared to what was in .trackNew.
// - set of revisions for which new entries in .vδT have been created.
//
// XXX place
func (δTtail *ΔTtail) rebuild(root zodb.Oid, δZtail *zodb.ΔTail, db *zodb.DB) (δtrackSet blib.PPTreeSubSet, δrevSet setTid, err error) {
	defer xerr.Contextf(&err, "ΔTtail<%s> rebuild", root)
	// XXX locking

	tracefΔBtail("\nRebuild %s @%s .. @%s\n", root, δZtail.Tail(), δZtail.Head())
	tracefΔBtail("trackNew:       %v\n", δTtail.trackNew)

	trackNew := δTtail.trackNew
	δTtail.trackNew = blib.PPTreeSubSet{}

	if len(trackNew) == 0 {
		return nil, nil, nil
	}

	δrevSet = setTid{}

	// clone vδT before modifying it
	// queries such as SliceByRootRev return slices of vδT and we do not
	// want to change data that is already returned to user.
	δTtail.vδT = vδTClone(δTtail.vδT)

	// go backwards and merge vδT <- treediff(lo..hi/trackNew)
	vδZ := δZtail.Data()
	for {
		δtkeycov := &blib.RangedKeySet{}     // all keys coming into tracking set during this lo<-hi scan
		trackNewCur := trackNew.Clone() // trackNew adjusted as of when going to i<- entry
		for i := len(vδZ)-1; i>=0; i-- {
			δZ := vδZ[i]

			// XXX dup wrt Update?

			var atPrev zodb.Tid
			if i > 0 {
				atPrev = vδZ[i-1].Rev
			} else {
				atPrev = δZtail.Tail()
			}

			δtrackNew, δtkeycov_, newRevEntry, err := δTtail.rebuild1(atPrev, δZ, trackNewCur, db)
			if err != nil {
				return nil, nil, err
			}

			trackNewCur.ApplyΔ(δtrackNew)
			δtkeycov.UnionInplace(δtkeycov_)
			if newRevEntry {
				δrevSet.Add(δZ.Rev)
			}

			// XXX update .lastRevOf
		}

		// an iteration closer to tail may turn out to add a key to the tracking set.
		// We have to recheck all entries newer that revision for changes to that key,
		// for example:
		//
		//              8        5*
		//             / \  <-  / \
		//            2   8    2*  7
		//
		//     here initial tracked set is 5*-2*. Going to earlier revision
		//     2'th keycov range is widen from [-∞,5) to [-∞,7), so 5*-7 in
		//     later revision have to be rechecked because 7 was added into
		//     tracking set.
		//
		// Implement this via restarting from head and cycling until
		// set of tracked keys does not grow anymore.
		if δtkeycov.Empty() {
			break
		}

		err := widenTrackNew(trackNew, δtkeycov, root, δZtail.Head(), db)
		if err != nil {
			return nil, nil, err
		}
	}

	return trackNew, δrevSet, nil
}

// widenTrackNew widens trackNew to cover δtkeycov.
// XXX -> widenTrackSet?
func widenTrackNew(trackNew blib.PPTreeSubSet, δtkeycov *blib.RangedKeySet, root zodb.Oid, at zodb.Tid, db *zodb.DB) (err error) {
	// XXX errctx, debug
	defer xerr.Contextf(&err, "widenTrackNew tree<%s> @%s +%s", root, at, δtkeycov)

	txn, ctx := transaction.New(context.TODO()) // XXX
	defer txn.Abort()

	zhead, err := db.Open(ctx, &zodb.ConnOptions{At: at});  /*X*/ if err != nil { return err }
	xtree, err := zgetNodeOrNil(ctx, zhead, root);          /*X*/ if err != nil { return err }
	if xtree == nil {
		// root deleted -> root node covers [-∞,∞)
		trackNew.AddPath([]zodb.Oid{root})
		return nil
	}
	tree := xtree.(*Tree) // must succeed	XXX better explicit panic?

	top := &nodeInRange{prefix: nil, keycov: blib.KeyRange{KeyMin, KeyMax}, node: tree}
	V   := rangeSplit{top}
	for _, r := range δtkeycov.AllRanges() {
		lo := r.Lo
		for {
			b, err := V.GetToLeaf(ctx, lo);  /*X*/ if err != nil { return err }
			trackNew.AddPath(b.Path())

			// continue with next right bucket until r coverage is complete
			if r.Hi_ <= b.keycov.Hi_ {
				break
			}
			lo = b.keycov.Hi_ + 1
		}
	}
	return nil
}


// rebuild1 rebuilds δT for single δZ.
//
// δtrackNew/δtkeycov represents how trackNew changes when going through `atPrev <- δZ.Rev` .
// newRevEntry indicates whether δZ.Rev was not there before in .vδT and new corresponding δT entry was created.
func (δTtail *ΔTtail) rebuild1(atPrev zodb.Tid, δZ zodb.ΔRevEntry, trackNew blib.PPTreeSubSet, db *zodb.DB) (δtrackNew *blib.ΔPPTreeSubSet, δtkeycov *blib.RangedKeySet, newRevEntry bool, err error) {
	defer xerr.Contextf(&err, "rebuild1 %s<-%s", atPrev, δZ.Rev)

	debugfΔBtail("\n  rebuild1 @%s <- @%s\n", atPrev, δZ.Rev)
	debugfΔBtail("  δZ:\t%v\n", δZ.Changev)
	debugfΔBtail("  trackNew:  %v\n", trackNew)
	defer func() {
		debugfΔBtail("-> δtrackNew:   %v\n", δtrackNew)
		debugfΔBtail("-> δtkeycov:    %v\n", δtkeycov)
		debugfΔBtail("-> newRevEntry: %v\n", newRevEntry)
		debugfΔBtail("\n\n")
	}()

	δZTC, δtopsByRoot := δZConnectTracked(δZ.Changev, trackNew)

	// skip opening DB connections if there is no change to this tree
	if len(δtopsByRoot) == 0 {
		return blib.NewΔPPTreeSubSet(), &blib.RangedKeySet{}, false, nil
	}

	if len(δtopsByRoot) != 1 {
		panicf("BUG: δtopsByRoot has > 1 entries: %v\ntrackNew: %v\nδZ: %v", δtopsByRoot, trackNew, δZ)
	}
	var root  zodb.Oid
	var δtops setOid
	for root_, δtops_ := range δtopsByRoot {
		root  = root_
		δtops = δtops_
	}


	// open ZODB connection corresponding to "current" and "prev" states
	txn, ctx := transaction.New(context.TODO()) // XXX
	defer txn.Abort()

	zconnPrev, err := db.Open(ctx, &zodb.ConnOptions{At: atPrev})
	if err != nil {
		return nil, nil, false, err
	}
	zconnCurr, err := db.Open(ctx, &zodb.ConnOptions{At: δZ.Rev})
	if err != nil {
		return nil, nil, false, err
	}

	// diff backwards curr -> prev
	δT, δtrack, δtkeycov, err := treediff(ctx, root, δtops, δZTC, trackNew, zconnCurr, zconnPrev)
	if err != nil {
		return nil, nil, false, err
	}

	debugfΔBtail("  -> root<%s>  δkv*: %v  δtrack*: %v  δtkeycov*: %v\n", root, δT, δtrack, δtkeycov)

	if len(δT) == 0 { // an object might be resaved without change
		return δtrack, δtkeycov, false, nil
	}

	// δTtail.vδT <- merge δT*
	l := len(δTtail.vδT)
	j := sort.Search(l, func(k int) bool {
		return δZ.Rev <= δTtail.vδT[k].Rev
	})
	if j == l || δTtail.vδT[j].Rev != δZ.Rev {
		newRevEntry = true
		δTcurr := ΔTree{Rev: δZ.Rev, ΔKV: map[Key]ΔValue{}}
		// insert(@j, δTcurr)
		δTtail.vδT = append(δTtail.vδT[:j],
				    append([]ΔTree{δTcurr},
				    δTtail.vδT[j:]...)...)
	}
	δTcurr := δTtail.vδT[j]

	for k, δv := range δT {
		// the diff was backward; δTtail entries are with diff forward
		δv.New, δv.Old = δv.Old, δv.New

		δv_, already := δTcurr.ΔKV[k]
		if already {
			if δv != δv_ {
				panicf("[%v] inconsistent δv:\nδTcurr: %v\nδT:  %v", k, δTcurr, δT)
			}
		} else {
			δTcurr.ΔKV[k] = δv
		}
	}

	// XXX update .lastRevOf	(here?)
	return δtrack, δtkeycov, newRevEntry, nil
}

// Update updates δB with per-object-level ZODB changes.
//
// Only those objects from δZ that belong to tracked set are guaranteed to be
// taken into account. In other words a tree history will assuredly include
// only those keys, that correspond to tracked subset of δZ.
//
// δZ must include all objects changed by ZODB transaction.
//
// TODO optionally accept zconnOld/zconnNew from client
func (δBtail *ΔBtail) Update(δZ *zodb.EventCommit) (_ ΔB, err error) {
	headOld := δBtail.Head()
	defer xerr.Contextf(&err, "ΔBtail.Update %s -> %s", headOld, δZ.Tid)

	δB1, err := δBtail._Update1(δZ)

	δB := ΔB{Rev: δZ.Tid, ΔByRoot: make(map[zodb.Oid]map[Key]ΔValue)}
	for root, δT1 := range δB1.ByRoot {
		δTtail := δBtail.vδTbyRoot[root] // must succeed

		// δtkeycov1 != ø   ->  rebuild δTtail with trackNew ~= δtkeycov1
		if !δT1.δtkeycov1.Empty() && δBtail.δZtail.Len() > 1 {
			trackNew := blib.PPTreeSubSet{}
			err := widenTrackNew(trackNew, δT1.δtkeycov1, root, δBtail.Head(), δBtail.db)
			if err != nil {
				return ΔB{}, err
			}
			// XXX assert δTtail.trackNew.Empty()
			δTtail.trackNew = trackNew

			// XXX vvv we can skip computing diff for HEAD~..HEAD (we just
			//     computed it in _Update1)? (or not - trackNew is as of @head ?)
			δtrackSet, δrevSet, err := δTtail.rebuild(root, δBtail.δZtail, δBtail.db)
			if err != nil {
				return ΔB{}, err
			}
			δBtail.trackSet.UnionInplace(δtrackSet)
			δBtail.vδBroots_Update(root, δrevSet)
		}

		// build δB. Even if δT=ø after _Update1, but δtkeycov1 != ø, above
		// rebuild could result in head δT becoming != ø. Detect that δTtail head
		// is anew by comparing to δZ.Rev.
		l := len(δTtail.vδT)
		if l > 0 {
			δT := δTtail.vδT[l-1] // δT head
			if δT.Rev == δZ.Tid {
				δB.ΔByRoot[root] = δT.ΔKV
			}
		}

		// XXX rebuild lastRevOf
	}

	// vδBroots += δB  (δB.Rev could be already there added by ^^^ rebuild)
	for root := range δB.ΔByRoot {
		δBtail.vδBroots_Update1(root, δB.Rev)
	}

	return δB, err
}

// _Update1 serves Update and performs direct update of δTtail head elements from δZ.
// On key coverage growth rebuilding tail of the history is done by Update itself.
//
// _Update1 is also used in tests to verify δtkeycov return from treediff.
type _ΔBUpdate1 struct {
	ByRoot map[zodb.Oid]*_ΔTUpdate1
}
type _ΔTUpdate1 struct {
	δtkeycov1 *blib.RangedKeySet   // {} root -> δtrackedKeys after first treediff (always grow)
	δtrack    *blib.ΔPPTreeSubSet  // XXX kill (not used)
}
func (δBtail *ΔBtail) _Update1(δZ *zodb.EventCommit) (δB1 _ΔBUpdate1, err error) {
	headOld := δBtail.Head()
	defer xerr.Contextf(&err, "ΔBtail.update1 %s -> %s", headOld, δZ.Tid)

	tracefΔBtail("\nUpdate  @%s -> @%s    δZ: %v\n", δBtail.Head(), δZ.Tid, δZ.Changev)
	tracefΔBtail("trackSet:       %v\n", δBtail.trackSet)
	for _, root := range δBtail.trackNewRoots.SortedElements() {
		δTtail := δBtail.vδTbyRoot[root]
		tracefΔBtail("[%s].trackNew:  %v\n", root, δTtail.trackNew)
	}

	δB1 = _ΔBUpdate1{ByRoot: make(map[zodb.Oid]*_ΔTUpdate1)}

	// update .trackSet and vδB from .trackNew
	err = δBtail.rebuildAll()
	if err != nil {
		return δB1, err
	}

	// XXX dup wrt rebuild?

	δBtail.δZtail.Append(δZ.Tid, δZ.Changev)

	δZTC, δtopsByRoot := δZConnectTracked(δZ.Changev, δBtail.trackSet)

	// skip opening DB connections if there is no change to any tree node
	if len(δtopsByRoot) == 0 {
		return δB1, nil
	}

	// open ZODB connections corresponding to "old" and "new" states
	// TODO caller should provide one of those (probably new) as usually it has it
	txn, ctx := transaction.New(context.TODO()) // XXX
	defer txn.Abort()
	zconnOld, err := δBtail.db.Open(ctx, &zodb.ConnOptions{At: headOld})
	if err != nil {
		return δB1, err
	}
	zconnNew, err := δBtail.db.Open(ctx, &zodb.ConnOptions{At: δZ.Tid})
	if err != nil {
		return δB1, err
	}

	for root, δtops := range δtopsByRoot {
		δT, δtrack, δtkeycov, err := treediff(ctx, root, δtops, δZTC, δBtail.trackSet, zconnOld, zconnNew)
		if err != nil {
			return δB1, err
		}

		tracefΔBtail("\n-> root<%s>  δkv: %v  δtrack: %v  δtkeycov: %v\n", root, δT, δtrack, δtkeycov)

		// XXX also needs vδT clone here?
		δTtail := δBtail.vδTbyRoot[root] // must be there
		if len(δT) > 0 { // an object might be resaved without change
			δTtail.vδT = append(δTtail.vδT, ΔTree{Rev: δZ.Tid, ΔKV: δT})
		}

		δBtail.trackSet.ApplyΔ(δtrack)
		δB1.ByRoot[root] = &_ΔTUpdate1{δtkeycov1: δtkeycov, δtrack: δtrack}
	}

	return δB1, nil
}

// vδBroots_Update updates .vδBroots to remember that ΔTtail for root has
// changed entries with δrevSet revisions.
//
// XXX place  TODO δrevSet -> []rev↑
func (δBtail *ΔBtail) vδBroots_Update(root zodb.Oid, δrevSet setTid) {
	// XXX locking
	for rev := range δrevSet {
		δBtail.vδBroots_Update1(root, rev)
	}
}

func (δBtail *ΔBtail) vδBroots_Update1(root zodb.Oid, rev zodb.Tid) {
	l := len(δBtail.vδBroots)
	j := sort.Search(l, func(k int) bool {
		return rev <= δBtail.vδBroots[k].Rev
	})
	if j == l || rev != δBtail.vδBroots[j].Rev {
		δBroots := _ΔBroots{Rev: rev, ΔRoots: setOid{}}
		// insert(@j, δBroots)
		δBtail.vδBroots = append(δBtail.vδBroots[:j],
					 append([]_ΔBroots{δBroots},
					 δBtail.vδBroots[j:]...)...)
	}
	δBroots := δBtail.vδBroots[j]
	δBroots.ΔRoots.Add(root)
}

// ForgetPast forgets history entries with revision ≤ revCut.
func (δBtail *ΔBtail) ForgetPast(revCut zodb.Tid) {
	// XXX locking

	δBtail.δZtail.ForgetPast(revCut) // XXX stub

	// go through vδBroots till revcut -> find which trees to trim -> trim ΔTtails.

	totrim := setOid{} // roots whose ΔTtail has changes ≤ revCut
	icut := 0
	for ; icut < len(δBtail.vδBroots); icut++ {
		δBroots := δBtail.vδBroots[icut]
		if δBroots.Rev > revCut {
			break
		}
		totrim.Update(δBroots.ΔRoots)
	}

	// vδBroots[:icut] should be forgotten
	δBtail.vδBroots = append([]_ΔBroots(nil), δBtail.vδBroots[icut:]...)

	// trim roots
	for root := range totrim {
		δTtail := δBtail.vδTbyRoot[root] // must be present
		δTtail.forgetPast(revCut)
	}
}

func (δTtail *ΔTtail) forgetPast(revCut zodb.Tid) {
	// XXX locking
	// XXX lastRevOf

	icut := 0
	for ; icut < len(δTtail.vδT); icut++ {
		if δTtail.vδT[icut].Rev > revCut {
			break
		}
	}

	// vδT[:icut] should be forgotten
	// NOTE clones vδT because queries return vδT aliases
	δTtail.vδT = append([]ΔTree(nil), δTtail.vδT[icut:]...)
}


// GetAt tries to retrieve root[key]@at from δBtail data.
//
// If δBtail has δB entry that covers root[key]@at, corresponding value
// (VDEL means deletion) and valueExact=true are returned. If δBtail data
// allows to determine revision of root[key]@at value, corresponding revision
// and revExact=true are returned. If revision of root[key]@at cannot be
// determined (rev=δBtail.Tail, revExact=false) are returned.
//
// If δBtail has no δB entry that covers root[key]@at, return is
//
//	(value=VDEL, valueExact=false, rev=δBtail.Tail, revExact=false)
//
// .rev and exact=true are returned:
//
//	(δB[root/key].δvalue.New, δB.rev, exact=true)
//
// If δBtail has no δB entry for root[key] with .rev ≤ @at, return is
//
//	(VDEL, δBtail.Tail, exact=false)
//
// key must be tracked
// at  must ∈ (tail, head]
func (δBtail *ΔBtail) GetAt(root zodb.Oid, key Key, at zodb.Tid) (value Value, rev zodb.Tid, valueExact, revExact bool, err error) {
	defer xerr.Contextf(&err, "δBtail: root<%s>: get %d @%s", root, key, at)
	// XXX key not tracked		-> panic

	tail := δBtail.Tail()
	head := δBtail.Head()
	if !(tail < at && at <= head) {
		panicf("at out of bounds: at: @%s,  (tail, head] = (@%s, @%s]", at, tail, head)
	}

	// XXX locking

	value = VDEL
	valueExact = false
	rev = tail
	revExact = false

	err = δBtail.rebuild1IfNeeded(root)
	if err != nil {
		return value, rev, valueExact, revExact, err
	}

	δTtail := δBtail.vδTbyRoot[root]
	if δTtail == nil {
		panicf("δBtail: root<%s> not tracked", root)
	}

	// XXX -> index lastXXXOf(key) | linear scan ↓ looking for change <= at
	for i := len(δTtail.vδT)-1; i >= 0; i-- {
		δT := δTtail.vδT[i]
		δvalue, ok_ := δT.ΔKV[key]
		if ok_ {
			valueExact = true
			if δT.Rev > at {
				value = δvalue.Old
			} else {
				value = δvalue.New
				rev = δT.Rev
				revExact = true
				break
			}
		}
	}

	return value, rev, valueExact, revExact, nil
}

// XXX don't need
//func (δBtail *ΔBtail) SliceByRev(lo, hi zodb.Tid) /*readonly*/ []ΔB {
//	xtail.AssertSlice(δBtail, lo, hi)
//	panic("TODO")
//}

// SliceByRootRev returns history of a tree changes in (lo, hi] range.
//
// it must be called with the following condition:
//
//	tail ≤ lo ≤ hi ≤ head
//
// the caller must not modify returned slice.
//
// Note: contrary to regular go slicing, low is exclusive while high is inclusive.
//
// XXX root -> *Tree ?
func (δBtail *ΔBtail) SliceByRootRev(root zodb.Oid, lo, hi zodb.Tid) /*readonly*/[]ΔTree {
	xtail.AssertSlice(δBtail, lo, hi)
	// XXX locking

	err := δBtail.rebuild1IfNeeded(root)
	if err != nil {
		panic(err) // XXX
	}

	δTtail, ok := δBtail.vδTbyRoot[root]
	if !ok {
		return []ΔTree{}
	}
	// XXX dup data - because they can be further rebuilt in parallel to caller using them
	// XXX no -> dup data in rebuild, not here

	vδT := δTtail.vδT
	l := len(vδT)
	if l == 0 {
		return nil
	}

	// find max j : [j].rev ≤ hi		XXX linear scan -> binary search
	j := l - 1
	for ; j >= 0 && vδT[j].Rev > hi; j-- {}
	if j < 0 {
		return nil // ø
	}

	// find max i : [i].rev > lo           XXX linear scan -> binary search
	i := j
	for ; i >= 0 && vδT[i].Rev > lo; i-- {}
        i++

	return vδT[i:j+1]
}


// ΔZtail returns raw ZODB changes accumulated in δBtail so far.
//
// the caller must not modify returned δZtail.
func (δBtail *ΔBtail) ΔZtail() /*readonly*/*zodb.ΔTail {
	return δBtail.δZtail
}

// DB returns database handle that δBtail is using to access ZODB.
func (δBtail *ΔBtail) DB() *zodb.DB {
	return δBtail.db
}


func tracefΔBtail(format string, argv ...interface{}) {
	if traceΔBtail {
		fmt.Printf(format, argv...)
	}
}

func debugfΔBtail(format string, argv ...interface{}) {
	if debugΔBtail {
		fmt.Printf(format, argv...)
	}
}
